package org.apache.catalina.ha.deploy;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicLong;
import org.apache.catalina.ha.deploy.FileMessage;
import org.apache.catalina.tribes.Member;
import org.apache.juli.logging.Log;
import org.apache.juli.logging.LogFactory;
import org.apache.tomcat.util.buf.HexUtils;
import org.apache.tomcat.util.res.StringManager;

public class FileMessageFactory {

   private static final Log log = LogFactory.getLog(FileMessageFactory.class);
   private static final StringManager sm = StringManager.getManager(FileMessageFactory.class);
   public static final int READ_SIZE = 10240;
   protected final File file;
   protected final boolean openForWrite;
   protected boolean closed = false;
   protected FileInputStream in;
   protected FileOutputStream out;
   protected int nrOfMessagesProcessed = 0;
   protected long size = 0L;
   protected long totalNrOfMessages = 0L;
   protected AtomicLong lastMessageProcessed = new AtomicLong(0L);
   protected final Map msgBuffer = new ConcurrentHashMap();
   protected byte[] data = new byte[10240];
   protected boolean isWriting = false;
   /** @deprecated */
   @Deprecated
   protected long creationTime = 0L;
   protected long lastModified = 0L;
   protected int maxValidTime = -1;


   private FileMessageFactory(File f, boolean openForWrite) throws FileNotFoundException, IOException {
      this.file = f;
      this.openForWrite = openForWrite;
      if(log.isDebugEnabled()) {
         log.debug("open file " + f + " write " + openForWrite);
      }

      if(openForWrite) {
         if(!this.file.exists() && !this.file.createNewFile()) {
            throw new IOException(sm.getString("fileNewFail", new Object[]{this.file}));
         }

         this.out = new FileOutputStream(f);
      } else {
         this.size = this.file.length();
         this.totalNrOfMessages = this.size / 10240L + 1L;
         this.in = new FileInputStream(f);
      }

      this.creationTime = System.currentTimeMillis();
      this.lastModified = System.currentTimeMillis();
   }

   public static FileMessageFactory getInstance(File f, boolean openForWrite) throws FileNotFoundException, IOException {
      return new FileMessageFactory(f, openForWrite);
   }

   public FileMessage readMessage(FileMessage f) throws IllegalArgumentException, IOException {
      this.checkState(false);
      int length = this.in.read(this.data);
      if(length == -1) {
         this.cleanup();
         return null;
      } else {
         f.setData(this.data, length);
         f.setTotalNrOfMsgs(this.totalNrOfMessages);
         f.setMessageNumber(++this.nrOfMessagesProcessed);
         return f;
      }
   }

   public boolean writeMessage(FileMessage msg) throws IllegalArgumentException, IOException {
      if(!this.openForWrite) {
         throw new IllegalArgumentException(sm.getString("fileMessageFactory.cannotWrite"));
      } else {
         if(log.isDebugEnabled()) {
            log.debug("Message " + msg + " data " + HexUtils.toHexString(msg.getData()) + " data length " + msg.getDataLength() + " out " + this.out);
         }

         if((long)msg.getMessageNumber() <= this.lastMessageProcessed.get()) {
            log.warn(sm.getString("fileMessageFactory.duplicateMessage", new Object[]{msg.getContextName(), msg.getFileName(), HexUtils.toHexString(msg.getData()), Integer.valueOf(msg.getDataLength())}));
            return false;
         } else {
            FileMessage previous = (FileMessage)this.msgBuffer.put(Long.valueOf((long)msg.getMessageNumber()), msg);
            if(previous != null) {
               log.warn(sm.getString("fileMessageFactory.duplicateMessage", new Object[]{msg.getContextName(), msg.getFileName(), HexUtils.toHexString(msg.getData()), Integer.valueOf(msg.getDataLength())}));
               return false;
            } else {
               this.lastModified = System.currentTimeMillis();
               FileMessage next = null;
               synchronized(this) {
                  if(this.isWriting) {
                     return false;
                  }

                  next = (FileMessage)this.msgBuffer.get(Long.valueOf(this.lastMessageProcessed.get() + 1L));
                  if(next == null) {
                     return false;
                  }

                  this.isWriting = true;
               }

               while(next != null) {
                  this.out.write(next.getData(), 0, next.getDataLength());
                  this.lastMessageProcessed.incrementAndGet();
                  this.out.flush();
                  if((long)next.getMessageNumber() == next.getTotalNrOfMsgs()) {
                     this.out.close();
                     this.cleanup();
                     return true;
                  }

                  synchronized(this) {
                     next = (FileMessage)this.msgBuffer.get(Long.valueOf(this.lastMessageProcessed.get() + 1L));
                     if(next == null) {
                        this.isWriting = false;
                     }
                  }
               }

               return false;
            }
         }
      }
   }

   public void cleanup() {
      if(this.in != null) {
         try {
            this.in.close();
         } catch (IOException var3) {
            ;
         }
      }

      if(this.out != null) {
         try {
            this.out.close();
         } catch (IOException var2) {
            ;
         }
      }

      this.in = null;
      this.out = null;
      this.size = 0L;
      this.closed = true;
      this.data = null;
      this.nrOfMessagesProcessed = 0;
      this.totalNrOfMessages = 0L;
      this.msgBuffer.clear();
      this.lastMessageProcessed = null;
   }

   protected void checkState(boolean openForWrite) throws IllegalArgumentException {
      if(this.openForWrite != openForWrite) {
         this.cleanup();
         if(openForWrite) {
            throw new IllegalArgumentException(sm.getString("fileMessageFactory.cannotWrite"));
         } else {
            throw new IllegalArgumentException(sm.getString("fileMessageFactory.cannotRead"));
         }
      } else if(this.closed) {
         this.cleanup();
         throw new IllegalArgumentException(sm.getString("fileMessageFactory.closed"));
      }
   }

   /** @deprecated */
   @Deprecated
   public static void main(String[] args) throws Exception {
      System.out.println("Usage: FileMessageFactory fileToBeRead fileToBeWritten");
      System.out.println("Usage: This will make a copy of the file on the local file system");
      FileMessageFactory read = getInstance(new File(args[0]), false);
      FileMessageFactory write = getInstance(new File(args[1]), true);
      FileMessage msg = new FileMessage((Member)null, args[0], args[0]);
      msg = read.readMessage(msg);
      if(msg == null) {
         System.out.println("Empty input file : " + args[0]);
      } else {
         System.out.println("Expecting to write " + msg.getTotalNrOfMsgs() + " messages.");

         int cnt;
         for(cnt = 0; msg != null; msg = read.readMessage(msg)) {
            write.writeMessage(msg);
            ++cnt;
         }

         System.out.println("Actually wrote " + cnt + " messages.");
      }
   }

   public File getFile() {
      return this.file;
   }

   public boolean isValid() {
      if(this.maxValidTime > 0) {
         long timeNow = System.currentTimeMillis();
         long timeIdle = (timeNow - this.lastModified) / 1000L;
         if(timeIdle > (long)this.maxValidTime) {
            this.cleanup();
            if(this.file.exists()) {
               if(this.file.delete()) {
                  log.warn(sm.getString("fileMessageFactory.delete", new Object[]{this.file, Long.toString((long)this.maxValidTime)}));
               } else {
                  log.warn(sm.getString("fileMessageFactory.deleteFail", new Object[]{this.file}));
               }
            }

            return false;
         }
      }

      return true;
   }

   public int getMaxValidTime() {
      return this.maxValidTime;
   }

   public void setMaxValidTime(int maxValidTime) {
      this.maxValidTime = maxValidTime;
   }

}
